@@ -1048,7 +1048,7 @@ void Torrent::Accept(int s,const sockaddr_u *addr,IOBuffer *rb)
 void Torrent::AddPeer(TorrentPeer *peer)
 {
    if(black_list->Listed(peer->GetAddress())) {
-      delete peer;
+      Delete(peer);
       return;
    }
    for(int i=0; i<peers.count(); i++) {
@@ -1056,7 +1056,7 @@ void Torrent::AddPeer(TorrentPeer *peer)
 	 if(peer->Connected() && !peers[i]->Connected()) {
 	    peers[i]=peer;
 	 } else {
-	    delete peer;
+	    Delete(peer);
 	 }
 	 return;
       }
@@ -1621,17 +1621,21 @@ const xstring& Torrent::Status()
 
    char dn_buf[LONGEST_HUMAN_READABLE + 1];
    char up_buf[LONGEST_HUMAN_READABLE + 1];
-   xstring& buf=xstring::format("dn:%s %sup:%s %scomplete:%u/%u (%u%%)",
+   xstring& buf=xstring::format("dn:%s %sup:%s %s",
       human_readable(total_recv, dn_buf, human_autoscale|human_SI, 1, 1),
       recv_rate.GetStrS(),
       human_readable(total_sent, up_buf, human_autoscale|human_SI, 1, 1),
-      send_rate.GetStrS(),
-      complete_pieces,total_pieces,
-      unsigned((total_length-total_left)*100/total_length));
-   if(end_game)
-      buf.append(" end-game");
-   buf.append(' ');
-   buf.append(recv_rate.GetETAStrFromSize(total_left));
+      send_rate.GetStrS());
+   if(!complete) {
+      buf.appendf("complete:%u/%u (%u%%)",complete_pieces,total_pieces,
+	 unsigned((total_length-total_left)*100/total_length));
+      buf.append(' ');
+      buf.append(recv_rate.GetETAStrFromSize(total_left));
+      if(end_game)
+	 buf.append(" end-game");
+   } else {
+      buf.appendf("complete, ratio:%f",GetRatio());
+   }
    return buf;
 }
 
@@ -3030,9 +3034,13 @@ const char *TorrentPeer::Status()
       buf.append("am-interested ");
    if(am_choking)
       buf.append("am-choking ");
-   if(parent->HasMetadata())
-      buf.appendf("complete:%u/%u (%u%%)",peer_complete_pieces,parent->total_pieces,
-	 peer_complete_pieces*100/parent->total_pieces);
+   if(parent->HasMetadata()) {
+      if(peer_complete_pieces<parent->total_pieces)
+	 buf.appendf("complete:%u/%u (%u%%)",peer_complete_pieces,parent->total_pieces,
+	    peer_complete_pieces*100/parent->total_pieces);
+      else
+	 buf.append("complete");
+   }
    return buf;
 }
 
@@ -3387,7 +3395,7 @@ void Torrent::Dispatch(const xstring& info_hash,int sock,const sockaddr_u *remot
       LogError(3,"peer %s sent unknown info_hash=%s in handshake",
 	 remote_addr->to_string(),info_hash.hexdump());
       close(sock);
-      delete recv_buf;
+      Delete(recv_buf);
       return;
    }
    t->Accept(sock,remote_addr,recv_buf);
@@ -3488,7 +3496,7 @@ xstring& TorrentJob::FormatStatus(xstring& s,int v,const char *tab)
    if(name)
       s.appendf("%sName: %s\n",tab,name);
    s.appendf("%s%s\n",tab,torrent->Status().get());
-   if(torrent->GetRatio()>0)
+   if(!torrent->Complete() && torrent->GetRatio()>0)
       s.appendf("%sratio: %f\n",tab,torrent->GetRatio());
 
    if(v>2) {
